/*
    Copyright (C) 2020 Daniel Schultz

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#include "fmpz_poly.h"
#include "fq.h"

void fq_get_fmpz_poly(fmpz_poly_t a, const fq_t b, const fq_ctx_t FLINT_UNUSED(ctx))
{
    fmpz_poly_set(a, b);
}
